import logging

from .utils import pretty

log = logging.getLogger(__name__)


# Alternative names: Definition
class Entity(object):
    ''' The base class for Greengrass model elements '''

    _session = None

    def __init__(self, group, state):
        self.type = "REPLACE WITH CHILD CLASS TYPE!"
        self._state = state
        self._group = group

        # Objects dependent on this, (immediate children),
        # thus must be removed prior to `this` object is removed.
        self._dependants = set()

        # Object types that must be created prior to this object's creation
        # Note that it's list of Types (classes), not instances.
        self._requirements = []

    def add_dependant(self, entity):
        self._dependants.add(entity)

    def create(self, update_group_version=True):
        if self._pre_create():
            self._do_create()
            self._post_create(update_group_version)

    def _pre_create(self):
        log.info("Creating {}...".format(self.type.lower()))

        if not self._group.get(self.type):
            log.warning("{} not defined, skipping.".format(self.type))
            return False

        if self._state.get(self.type):
            log.warning("{} already created. Remove before creating again.".format(self.type))
            return False

        # Check if requred objects alrady created, if not - warn and exit.
        missed = [r for r in self._requirements if r not in self._state.entities()]
        if missed:
            log.warning("{} must be created before {}. Please create first.".format(
                ', '.join(missed), self.type))
            return False

        return True

    def _do_create(self):
        raise NotImplementedError

    def _post_create(self, update_group_version):
        if update_group_version:
            log.info("Updating group version with new {}...".format(self.type))
            self.create_group_version(self._state)

        log.info("{} created OK!".format(self.type))

    def remove(self, update_group_version=True):
        if self._pre_remove():
            self._do_remove()
            self._post_remove(update_group_version)

    def _pre_remove(self):
        log.info("Removing {}".format(self.type.lower()))

        if not self._state.get(self.type):
            log.warning("{} does not exist, skipping remove.".format(self.type))
            return False

        return True

    def _do_remove(self):
        raise NotImplementedError

    def _post_remove(self, update_group_version):
        if update_group_version:
            log.info("Updating group version with {} removed...".format(self.type))
            self.create_group_version(self._state)

        self._state.remove(self.type)
        log.info("{} removed OK!".format(self.type.lower()))

    @classmethod
    def create_group_version(klass, state):
        if not state.get('Group.Id'):
            log.debug("Attempting to update group version but the group doesn't exist, skipping!")
            return

        # Compile a list of non-empty arguments from the super-set,
        # https://docs.aws.amazon.com/greengrass/latest/apireference/definitions-groupversion.html
        kwargs = dict(
            GroupId=state.get('Group')['Id'],  # REQUIRED
            CoreDefinitionVersionArn=state.get(
                'CoreDefinition.LatestVersionArn', ""),
            FunctionDefinitionVersionArn=state.get(
                'Lambdas.FunctionDefinition.LatestVersionArn', ""),
            SubscriptionDefinitionVersionArn=state.get(
                'Subscriptions.LatestVersionArn', ""),
            LoggerDefinitionVersionArn=state.get(
                'Loggers.LatestVersionArn', ""),
            ResourceDefinitionVersionArn=state.get(
                'Resources.LatestVersionArn', ""),
            ConnectorDefinitionVersionArn=state.get(
                'Connectors.LatestVersionArn', ""),
            DeviceDefinitionVersionArn=""  # NOT IMPLEMENTED
        )
        args = dict((k, v) for k, v in kwargs.items() if v)
        log.debug("Creating group version with settings:\n{0}".format(pretty(args)))

        _gg = Entity._session.client("greengrass")
        group_ver = _gg.create_group_version(**args)

        state.update('Group.Version', group_ver)



import logging
import json
import time
from botocore.exceptions import ClientError

from .entity import Entity
from .utils import rinse, mkdir, save_keys

log = logging.getLogger(__name__)


class Group(Entity):

    def __init__(self, group, state):
        super(Group, self).__init__(group, state)
        self.type = 'Group'
        self.name = group['Group']['name']

        self._region = Entity._session.region_name
        self._gg = Entity._session.client("greengrass")
        self._iot = Entity._session.client("iot")
        # self._lambda = s.client("lambda")
        # self._iam = s.client("iam")
        self._iot_endpoint = self._iot.describe_endpoint()['endpointAddress']

    def _do_create(self):
        log.info("Creating group '{}'".format(self._group['Group']['name']))

        g = rinse(self._gg.create_group(Name=self.name))
        self._state.update(self.type, g)

        self._create_cores()

    def _do_remove(self):

        log.info("Reseting deployments forcefully, if they exist")
        # print(json.dumps(self._state._state['Group'], indent=4, default=str))
        self._gg.reset_deployments(GroupId=self._state.get('Group.Id'), Force=True)

        self._remove_cores()

        log.info("Deleting group '{0}'".format(self._state.get('Group.Id')))
        self._gg.delete_group(GroupId=self._state.get('Group.Id'))

    def _create_cores(self):
        # Design notes:
        # 1) In practice one Group has only one Core. But AWS model implies one-to-many relations.
        #    I go for the model.
        # 2) It might be proper to make Cores a separate Entity. But Group and Core are tightly
        #    coupled, always created/removed together and can't be used independent of each other.
        #    Thus I keep them in the same class.
        cores = []
        initial_version = {'Cores': []}

        for core in self._group['Cores']:
            try:
                name = core['name']
                log.info("Creating a thing for core {0}".format(name))
                keys_cert = rinse(self._iot.create_keys_and_certificate(setAsActive=True))
                core_thing = rinse(self._iot.create_thing(thingName=name))

                # Attach the previously created Certificate to the created Thing
                self._iot.attach_thing_principal(
                    thingName=name, principal=keys_cert['certificateArn'])
                policy = self._create_and_attach_thing_policy(
                    thing_name=name,
                    policy_doc=self._get_core_policy_doc(),
                    thing_cert_arn=keys_cert['certificateArn']
                )

                cores.append({
                    'name': name,
                    'thing': core_thing,
                    'keys': keys_cert,
                    'policy': policy
                })

                self._state.update('Cores', cores)

                initial_version['Cores'].append({
                    'Id': name,
                    'CertificateArn': keys_cert['certificateArn'],
                    'SyncShadow': core['SyncShadow'],
                    'ThingArn': core_thing['thingArn']
                })

                save_keys(core['key_path'], name, keys_cert)

                self._create_ggc_config_file(core['config_path'], "config.json", core_thing)

            except Exception as e:
                log.error("Error creating core {0}: {1}".format(name, str(e)))
                raise

            log.debug("Creating Core definition with InitialVersion={0}".format(
                initial_version))

            core_def = rinse(self._gg.create_core_definition(
                Name="{0}_core_def".format(self._group['Group']['name']),
                InitialVersion=initial_version
            ))

            self._state.update('CoreDefinition', core_def)

            log.info("Created Core definition Arn:{0} Id:{1}".format(
                core_def['Arn'], core_def['Id']))

    def _create_and_attach_thing_policy(self, thing_name, policy_doc, thing_cert_arn):
        try:
            policy_name = "{0}-policy".format(thing_name)
            policy = rinse(self._iot.create_policy(
                policyName=policy_name,
                policyDocument=policy_doc)
            )
        except ClientError as ce:
            if ce.response['Error']['Code'] == 'EntityAlreadyExists':
                log.warning(
                    "Policy '{0}' exists. Using existing Policy".format(policy_name))
            else:
                log.error("Unexpected Error: {0}".format(ce))
                raise

        self._iot.attach_principal_policy(
            policyName=policy_name,
            principal=thing_cert_arn
        )
        log.info("Created policy {0} for {1} and attached to certificate {2}".format(
            policy_name, thing_name, thing_cert_arn))

        return policy

    def _get_core_policy_doc(self):
        # TODO: redo as template and read from definition file
        core_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "iot:Publish",
                        "iot:Subscribe",
                        "iot:Connect",
                        "iot:Receive",
                        "iot:GetThingShadow",
                        "iot:DeleteThingShadow",
                        "iot:UpdateThingShadow"
                    ],
                    "Resource": ["arn:aws:iot:" + self._region + ":*:*"]
                },
                {
                    "Effect": "Allow",
                    "Action": [
                        "greengrass:AssumeRoleForGroup",
                        "greengrass:CreateCertificate",
                        "greengrass:GetConnectivityInfo",
                        "greengrass:GetDeployment",
                        "greengrass:GetDeploymentArtifacts",
                        "greengrass:UpdateConnectivityInfo",
                        "greengrass:UpdateCoreDeploymentStatus"
                    ],
                    "Resource": ["*"]
                }
            ]
        }
        return json.dumps(core_policy)

    def _create_ggc_config_file(self, path, name, core_thing):

        log.info("Creating GGC config file with core {0} at {1}/{2}".format(
            core_thing['thingName'], path, name))

        config = {
            "system": {
                "shadowSyncTimeout": 10
            },
            "coreThing": {
                "caPath": "root.ca.pem",
                "certPath": core_thing['thingName'] + ".pem",
                "keyPath": core_thing['thingName'] + ".key",
                "thingArn": core_thing['thingArn'],
                "iotHost": self._iot_endpoint,
                "ggHost": "greengrass.iot." + self._region + ".amazonaws.com",
                "keepAlive": 600
            },
            "runtime": {
                "cgroup": {
                    "useSystemd": "yes"
                }
            },
            "managedRespawn": False
        }

        mkdir(path)
        with open(path + '/' + name, 'w') as f:
            json.dump(config, f, indent=4, separators=(',', ' : '))

    def _remove_cores(self):

        for core in self._state.get('Cores'):
            try:
                thing_name = core['thing']['thingName']
                cert_id = core['keys']['certificateId']
                log.info("Removing core thing '{0}'' from core '{1}'".format(
                    thing_name, core['name']))

                log.debug("--- detaching policy: '{0}'".format(core['policy']['policyName']))
                self._iot.detach_principal_policy(
                    policyName=core['policy']['policyName'], principal=core['keys']['certificateArn'])

                log.debug("--- deleting policy: '{0}'".format(core['policy']['policyName']))
                self._iot.delete_policy(policyName=core['policy']['policyName'])

                log.debug("--- deactivating certificate: '{0}'".format(core['keys']['certificateId']))
                self._iot.update_certificate(
                    certificateId=cert_id, newStatus='INACTIVE')

                log.debug(
                    "--- detaching certificate '{0}' from thing '{1}'".format(cert_id, thing_name))
                self._iot.detach_thing_principal(
                    thingName=thing_name, principal=core['keys']['certificateArn'])

                log.debug(
                    "--- taking 1 sec. nap as a superstition to let certificate update propagate...")
                time.sleep(1)

                log.debug("--- deleting certificate: '{0}'".format(core['keys']['certificateId']))
                self._iot.delete_certificate(certificateId=core['keys']['certificateId'])

                log.debug("--- deleting thing: '{0}'".format(core['thing']['thingName']))
                self._iot.delete_thing(thingName=core['thing']['thingName'])

            except Exception as e:
                log.error("Error removing core {0}: {1}".format(core['name'], str(e)))

        self._state.remove('Cores')
        log.info("Cores removed OK!")

        core_def = self._state.get('CoreDefinition')
        log.info("Removing core definition '{0}'".format(core_def['Name']))
        self._gg.delete_core_definition(CoreDefinitionId=core_def['Id'])

        self._state.remove('CoreDefinition')
        log.info("Core definition '{0}' removed OK!".format(core_def['Name']))



import json
from .utils import rinse
from .entity import Entity


class Shadow(Entity):

    def __init__(self, group, state):
        super(Shadow, self).__init__(group, state)
        self.type = 'Shadow'
        self.name = group['Group']['name'] + '_shadow'

        self._requirements = ['Group']
        self._iot_data_plane = Entity._session.client("iot-data")

    def _do_create(self):
        exists = os.path.isfile('./initShadow.json')
        if exists:
            f = open('./initShadow.json', encoding='utf-8')
            text = f.read()    # unicode, not bytes
            response = self._iot_data_plane.update_thing_shadow(
                thingName=self._group['Cores'][0]['name'],
                payload=text
            )
        else:
            print("No file named initShadow.json defined !")

    def _do_remove(self):
        # to do remove dns entry
        return
        # message = {
        #   'hostname': self._group['Cores'][0]['name']+'.dyn.doop.co.nz.', 
        #   'record_type': 'A',
        #   'ttl': 60,
        #   'shared_secret': self._group['DNS']['secret'],
        #   'lock_record': False, 
        #   'read_privilege': False, 
        #   'allow_internal': True,
        # }

        # response =  self._iot_data_plane.publish(
        #     topic='d1/dnsregister/'+self._group['Cores'][0]['name'],
        #     payload= json.dumps(message)
        # )
        # self._state.update('DNS.config', message)



import os
import fire
import yaml
import logging
import time
import uuid
from boto3 import session

from .__init__ import __version__

from .utils import rinse
from .entity import Entity
from .state import State
from .group import Group
from .lambdas import Lambdas
from .elambdas import ELambdas
from .subscriptions import Subscriptions
from .resources import Resources
from .loggers import Loggers
from .dns import Dns
from .dynamoEntry import DynamoEntry
from .shadow import Shadow

log = logging.getLogger(__name__)

DEFINITION_FILE = 'greengo.yaml'
MAGIC_DIR = '.gg'
STATE_FILE = os.path.join(MAGIC_DIR, 'gg_state.json')

DEPLOY_TIMEOUT = 90  # Timeout, seconds


class Commands(object):
    def __init__(self):

        s = session.Session()
        self._region = s.region_name
        if not self._region:
            log.error("AWS credentials and region must be setup. "
                      "Refer AWS docs at https://goo.gl/JDi5ie")
            exit(-1)

        log.info("AWS credentials found for region '{}'".format(self._region))

        Entity._session = s
        self._session = s
        # self._iot_endpoint = s.client("iot").describe_endpoint()['endpointAddress']

        try:
            with open(DEFINITION_FILE, 'r') as f:
                self.group = yaml.safe_load(f)
        except IOError:
            log.error("Group definition file `greengo.yaml` not found. "
                      "Create file, and define the group definition first. "
                      "See https://github.com/greengo for details.")
            exit(-1)

        self.state = State(STATE_FILE)
        try:
            self.name = self.group['Group']['name']
        except KeyError:   
            coreId = str(uuid.uuid4()) 
            self.group['DNS'] = {}
            self.group['DNS']['secret'] = str(uuid.uuid4())
            self.group['Group'] = {}
            self.group['Group']['name'] = 'G-'+coreId
            self.group['Cores'][0]['name'] = coreId
            self.group['Shadow'] = {}
            self.group['Shadow']['dev'] = True
            self.name = self.group['Group']['name']
            with open(DEFINITION_FILE, 'w') as f:
                yaml.dump(self.group, f)
                
    def version(self):
        print('Greengo version {}'.format(__version__))

    def core(self):
        print('Core ID {}'.format(self.group['Cores'][0]['name']))

    def state(self):
        print(self.state)

    def create(self):
        if self.state.get():
            log.error("Previously created group exists. Remove before creating!")
            return False

        log.info("[BEGIN] creating group {0}".format(self.group['Group']['name']))

        Group(self.group, self.state).create(update_group_version=False)

        # Create other entities, as defined in a group.

        Lambdas(self.group, self.state).create(update_group_version=False)
        
        ELambdas(self.group, self.state).create(update_group_version=False)

        Subscriptions(self.group, self.state).create(update_group_version=False)

        Resources(self.group, self.state).create(update_group_version=False)

        Loggers(self.group, self.state).create(update_group_version=False)

        Dns(self.group, self.state).create(update_group_version=False)

        DynamoEntry(self.group, self.state).create(update_group_version=False)

        Shadow(self.group, self.state).create(update_group_version=False)
        # Other entities go here...

        Group.create_group_version(self.state)

        log.info("[END] creating group {0}".format(self.group['Group']['name']))

    def remove(self):
        if not self.state.get():
            log.info("There seem to be nothing to remove.")
            return False

        log.info("[BEGIN] removing group {0}".format(self.group['Group']['name']))

        Subscriptions(self.group, self.state).remove(update_group_version=False)

        Lambdas(self.group, self.state).remove(update_group_version=False)

        ELambdas(self.group, self.state).remove(update_group_version=False)

        Resources(self.group, self.state).remove(update_group_version=False)

        Loggers(self.group, self.state).remove(update_group_version=False)

        DynamoEntry(self.group, self.state).remove(update_group_version=False)

        Shadow(self.group, self.state).remove(update_group_version=False)
        # Remove other entities here, before removing Group.

        Group(self.group, self.state).remove(update_group_version=False)

        self.state.remove()

        log.info("[END] removing group {0}".format(self.group['Group']['name']))

    def create_group(self):
        Group(self.group, self.state).create()

    def remove_group(self):
        Group(self.group, self.state).remove()

    def create_lambdas(self):
        Lambdas(self.group, self.state).create()
        ELambdas(self.group, self.state).create()

    def remove_lambdas(self):
        Lambdas(self.group, self.state).remove()
        ELambdas(self.group, self.state).remove()

    def create_subscriptions(self):
        Subscriptions(self.group, self.state).create()

    def remove_subscriptions(self):
        Subscriptions(self.group, self.state).remove()

    def update_lambda(self, lambda_name):
        Lambdas(self.group, self.state).update_lambda(lambda_name)

    def create_dynamoEntry(self):
        DynamoEntry(self.group, self.state).create()

    def create_resources(self):
        Resources(self.group, self.state).create(update_group_version=False)
        Group.create_group_version(self.state)

    def remove_resources(self):
        Resources(self.group, self.state).remove()
        
        
    def deploy(self):
        if not self.state:
            log.info("There is nothing to deploy. Do create first.")
            return

        log.info("Deploying group '{0}'".format(self.state.get('Group.Name')))
        gg = self._session.client("greengrass")
        deployment = gg.create_deployment(
            GroupId=self.state.get('Group.Id'),
            GroupVersionId=self.state.get('Group.Version.Version'),
            DeploymentType="NewDeployment")
        self.state.update('Deployment', rinse(deployment))

        for i in range(DEPLOY_TIMEOUT // 2):
            time.sleep(2)
            deployment_status = gg.get_deployment_status(
                GroupId=self.state.get('Group.Id'),
                DeploymentId=deployment['DeploymentId'])

            status = deployment_status.get('DeploymentStatus')

            log.debug("--- deploying... status: {0}".format(status))
            # Known status values: ['Building | InProgress | Success | Failure']
            if status == 'Success':
                log.info("--- SUCCESS!")
                self.state.update('Deployment.Status', rinse(deployment_status))
                return
            elif status == 'Failure':
                log.error("--- ERROR! {0}".format(deployment_status['ErrorMessage']))
                self.state.update('Deployment.Status', rinse(deployment_status))
                return

        log.warning(
            "--- Gave up waiting for deployment. Please check the status later. "
            "Make sure GreenGrass Core is running, connected to network, "
            "and the certificates match.")


def main():
    logging.basicConfig(
        format='%(asctime)s|%(name).10s|%(levelname).5s: %(message)s',
        level=logging.WARNING)
    logging.getLogger('greengo').setLevel(logging.DEBUG)
    logging.getLogger('__main__').setLevel(logging.DEBUG)  # There is a beter way...
    fire.Fire(Commands)

if __name__ == '__main__':
    main()



import os
import errno
import yaml
import logging

log = logging.getLogger(__name__)


def mkdir(path):
    ''' Create directory with entier path, like `mkdir -p`. '''
    try:
        os.makedirs(path)
    except OSError as exc:
        if not (exc.errno == errno.EEXIST and os.path.isdir(path)):
            raise


def rinse(boto_response):
    ''' Remove ResponseMetadata block from Boto / AWS API respone. '''
    boto_response.pop('ResponseMetadata', None)
    return boto_response


def pretty(d):
    """ Pretty object as YAML. """
    return yaml.safe_dump(d, default_flow_style=False)


def save_keys(path, name, keys_cert):
    # if not os.path.isfile(self.group['certs']['key_path']+"/root-CA.crt"):
    try:
        path = path + '/' if not path.endswith('/') else path
        mkdir(path)
        certname = path + name + ".pem"
        public_key_file = path + name + ".pub"
        private_key_file = path + name + ".key"
        with open(certname, "w") as pem_file:
            pem = keys_cert['certificatePem']
            pem_file.write(pem)
            log.info("Thing Name: {0}; PEM certifiate file: {1}".format(name, certname))

        with open(public_key_file, "w") as pub_file:
            pub = keys_cert['keyPair']['PublicKey']
            pub_file.write(pub)
            log.info("Thing Name: {0}; Public Key File: {1}".format(name, public_key_file))

        with open(private_key_file, "w") as prv_file:
            prv = keys_cert['keyPair']['PrivateKey']
            prv_file.write(prv)
            log.info("Thing Name: {0}; Private Key File: {1}".format(name, private_key_file))

    except OSError as e:
        log.error('Error while writing an certificate files. {0}'.format(e))
    except KeyError as e:
        log.error('Error while writing an certificate files. {0}'
                  'Check the keys {1}'.format(e, keys_cert))



import logging

from .utils import pretty, rinse
from .entity import Entity

log = logging.getLogger(__name__)


class Subscriptions(Entity):

    def __init__(self, group, state):
        super(Subscriptions, self).__init__(group, state)
        self.type = 'Subscriptions'
        self.name = group['Group']['name'] + '_subscriptions'

        self._requirements = ['Group']
        self._gg = Entity._session.client("greengrass")

    def _do_create(self):
        log.debug("Preparing subscription list...")
        subs = []
        for i, s in enumerate(self._group['Subscriptions']):
            log.debug("Subscription '{0}' - '{1}': {2}->{3}'".format(
                i, s['Subject'], s['Source'], s['Target']))
            if s['Callable']:
                subjectS = s['Subject']+'/{0}'.format(self._group['Cores'][0]['name'])
            else:
                subjectS = s['Subject']
            subs.append({
                'Id': str(i),
                'Source': self._resolve_subscription_destination(s['Source']),
                'Target': self._resolve_subscription_destination(s['Target']),
                'Subject': subjectS
            })
        log.debug("Subscription list is ready:\n{0}".format(pretty(subs)))

        log.info("Creating subscription definition: '{0}'".format(self.name))
        sub_def = rinse(self._gg.create_subscription_definition(
            Name=self.name,
            InitialVersion={'Subscriptions': subs}
        ))

        self._state.update('Subscriptions', sub_def)

        sub_def_ver = rinse(self._gg.get_subscription_definition_version(
            SubscriptionDefinitionId=self._state.get('Subscriptions.Id'),
            SubscriptionDefinitionVersionId=self._state.get('Subscriptions.LatestVersion')
        ))
        self._state.update('Subscriptions.LatestVersionDetails', sub_def_ver)

    def _do_remove(self):
        log.debug("Deleting subscription definition '{0}' Id='{1}".format(
            self.name, self._state.get('Subscriptions.Id')))
        self._gg.delete_subscription_definition(
            SubscriptionDefinitionId=self._state.get('Subscriptions.Id'))

    def _resolve_subscription_destination(self, d):
        p = [x.strip() for x in d.split('::')]
        if p[0] == 'cloud':
            return p[0]
        elif p[0] == 'Lambda':
            return self._lookup_lambda_qualified_arn(p[1])
        elif p[0] == 'Device':
            return self._lookup_device_arn(p[1])
        elif p[0] == 'GGShadowService':
            return p[0]
        elif p[0] == 'Connector':
            return self._lookup_connector_arn(p[1])
        else:
            raise ValueError(
                "Error parsing subscription destination '{0}'. Allowed values: "
                "'Lambda::', 'Device::', 'Connector::', 'GGShadowService', or 'cloud'.".format(d))

    def _lookup_lambda_qualified_arn(self, name):
        for l in self._state.get('Lambdas.FunctionDefinition.LatestVersionDetails.Definition.Functions', []):
            if l['Id'] == name:
                return l['FunctionArn']
        log.error("Lambda '{0}' not found".format(name))
        return None

    def _lookup_device_arn(self, name):
        raise NotImplementedError("WIP: Devices not implemented yet.")

    def _lookup_connector_arn(self, name):
        for l in self._state.get('Connectors.LatestVersionDetails.Definition.Connectors'):
            if l['Id'] == name:
                return l['ConnectorArn']
        log.error("Connector '{0}' not found".format(name))
        return None



import logging

from .utils import rinse
from .entity import Entity

log = logging.getLogger(__name__)


class Loggers(Entity):

    def __init__(self, group, state):
        super(Loggers, self).__init__(group, state)
        self.type = 'Loggers'
        self.name = group['Group']['name'] + '_loggers'

        self._requirements = ['Group']
        self._gg = Entity._session.client("greengrass")

    def _do_create(self):
        log.info("Creating loggers definition: '{0}'".format(self.name))

        l_def = rinse(self._gg.create_logger_definition(
            Name=self.name,
            InitialVersion={'Loggers': self._group['Loggers']}
        ))

        self._state.update('Loggers', l_def)

        l_def_ver = rinse(self._gg.get_logger_definition_version(
            LoggerDefinitionId=self._state.get('Loggers.Id'),
            LoggerDefinitionVersionId=self._state.get('Loggers.LatestVersion')
        ))

        self._state.update('Loggers.LatestVersionDetails', l_def_ver)

    def _do_remove(self):
        log.debug("Deleting logger definition '{0}' Id='{1}".format(
            self.name, self._state.get('Loggers.Id')))

        self._gg.delete_logger_definition(
            LoggerDefinitionId=self._state.get('Loggers.Id'))

        self._state.remove(self.type)



import shutil
import os
from time import sleep
import json
from botocore.exceptions import ClientError
import logging

from .entity import Entity
from .utils import pretty, rinse

MAGIC_DIR = '.gg'  # XXX: better way?

log = logging.getLogger(__name__)


class ELambdas(Entity):
    def __init__(self, group, state):
        super(ELambdas, self).__init__(group, state)
        self.type = 'ExistingLambdas'
        self.name = group['Group']['name'] + '_elambdas'

        self._requirements = ['Group']
        self._gg = Entity._session.client("greengrass")
        self._iam = Entity._session.client("iam")
        self._lambda = Entity._session.client("lambda")

    def _do_create(self):
        functions = []
        for l in self._group['ExistingLambdas']:
            functions.append({
                'Id': l['name'],
                'FunctionArn': l['aliasarn'],
                'FunctionConfiguration': l['greengrassConfig']
            })

        log.debug("Function definition list ready:\n{0}".format(pretty(functions)))

        log.info("Creating function definition: '{0}'".format(self.name + '_func_def_1'))
        fd = self._gg.create_function_definition(
            Name=self.name + '_func_def_1',
            InitialVersion={'Functions': functions}
        )
        self._state.update('Lambdas.FunctionDefinition', rinse(fd))

        fd_ver = self._gg.get_function_definition_version(
            FunctionDefinitionId=self._state.get('Lambdas.FunctionDefinition.Id'),
            FunctionDefinitionVersionId=self._state.get('Lambdas.FunctionDefinition.LatestVersion'))

        self._state.update('Lambdas.FunctionDefinition.LatestVersionDetails', rinse(fd_ver))

    def _do_remove(self):

        if not self._state.get('Lambdas.FunctionDefinition'):
            log.warning("Function definition was not created. Moving on...")
        else:
            fd_name = self._state.get('Lambdas.FunctionDefinition.Name')
            fd_id = self._state.get('Lambdas.FunctionDefinition.Id')
            log.info("Deleting function definition '{0}' Id='{1}".format(fd_name, fd_id))
            self._gg.delete_function_definition(FunctionDefinitionId=fd_id)
            self._state.remove('Lambdas.FunctionDefinition')

    def _default_lambda_role_arn(self):
        # TODO(XXX): Refactor, merge with _create_default_lambda_role;
        #            consider not messing with state here, move it up.
        if self._state.get('Lambdas.LambdaRole'):
            log.info("Default lambda role '{0}' already creted, RoleId={1} ".format(
                self._LAMBDA_ROLE_NAME, self._state.get('LambdaRole.Role.RoleId')))
        else:
            try:
                role = self._create_default_lambda_role()
            except ClientError as e:
                if e.response['Error']['Code'] == 'EntityAlreadyExists':
                    role = self._iam.get_role(RoleName=self._LAMBDA_ROLE_NAME)
                    log.warning("Role {0} already exists, reusing.".format(self._LAMBDA_ROLE_NAME))
                else:
                    raise e

            self._state.update('Lambdas.LambdaRole', rinse(role))
        return self._state.get('Lambdas.LambdaRole.Role.Arn')

    def _create_default_lambda_role(self):
        # TODO: redo as template and read from definition .yaml
        log.info("Creating default lambda role '{0}'".format(self._LAMBDA_ROLE_NAME))
        role_policy_document = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {
                        "Service": "lambda.amazonaws.com"
                    },
                    "Action": "sts:AssumeRole"

                }
            ]
        }

        role = self._iam.create_role(
            RoleName=self._LAMBDA_ROLE_NAME,
            AssumeRolePolicyDocument=json.dumps(role_policy_document)
        )

        log.info("Creating lambda role policy '{0}'".format(
            self._LAMBDA_ROLE_NAME + "_Policy"))
        inline_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "logs:CreateLogGroup",
                        "logs:CreateLogStream",
                        "logs:PutLogEvents"
                    ],
                    "Resource": "arn:aws:logs:*:*:*"
                }
            ]
        }

        self._iam.put_role_policy(
            RoleName=self._LAMBDA_ROLE_NAME,
            PolicyName=self._LAMBDA_ROLE_NAME + "_Policy",
            PolicyDocument=json.dumps(inline_policy))

        return role

    def _remove_default_lambda_role(self):
        for p in self._iam.list_role_policies(RoleName=self._LAMBDA_ROLE_NAME)['PolicyNames']:
            log.info("Deleting lambda role policy '{0}'".format(p))
            self._iam.delete_role_policy(RoleName=self._LAMBDA_ROLE_NAME, PolicyName=p)

        log.info("Deleting default lambda role '{0}'".format(self._LAMBDA_ROLE_NAME))
        self._iam.delete_role(RoleName=self._LAMBDA_ROLE_NAME)



import json
from .utils import rinse
from .entity import Entity


class Dns(Entity):

    def __init__(self, group, state):
        super(Dns, self).__init__(group, state)
        self.type = 'DNS'
        self.name = group['Group']['name'] + '_dns'

        self._requirements = ['Group']
        self._iot_data_plane = Entity._session.client("iot-data")

    def _do_create(self):
        message = {
          'hostname': self._group['Cores'][0]['name']+'.dyn.doop.co.nz.', 
          'record_type': 'A',
          'ttl': 60,
          'shared_secret': self._group['DNS']['secret'],
          'lock_record': False, 
          'read_privilege': False, 
          'allow_internal': True,
        }

        response =  self._iot_data_plane.publish(
            topic='d1/dnsregister/'+self._group['Cores'][0]['name'],
            payload= json.dumps(message)
        )
        self._state.update('DNS.config', message)


    def _do_remove(self):
        # to do remove dns entry
        return
        # message = {
        #   'hostname': self._group['Cores'][0]['name']+'.dyn.doop.co.nz.', 
        #   'record_type': 'A',
        #   'ttl': 60,
        #   'shared_secret': self._group['DNS']['secret'],
        #   'lock_record': False, 
        #   'read_privilege': False, 
        #   'allow_internal': True,
        # }

        # response =  self._iot_data_plane.publish(
        #     topic='d1/dnsregister/'+self._group['Cores'][0]['name'],
        #     payload= json.dumps(message)
        # )
        # self._state.update('DNS.config', message)



import json
from .utils import rinse
from .entity import Entity


class DynamoEntry(Entity):

    def __init__(self, group, state):
        super(DynamoEntry, self).__init__(group, state)
        self.type = 'DynamoEntry'
        self.name = group['Group']['name'] + '_dynamoEntry'

        self._requirements = ['Group']
        self._iot_data_plane = Entity._session.client("iot-data")

    def _do_create(self):
        message = {
            "device":self._group['Cores'][0]['name'],
            "type":"D1",
            "var1":"weightF",
            "var2":"waterL",
            "var3":"bay",
            "var4":"weightB"
        }

        response =  self._iot_data_plane.publish(
            topic='d1/register/'+self._group['Cores'][0]['name'],
            payload= json.dumps(message)
        )
        self._state.update('DynamoEntry.config', message)


    def _do_remove(self):
        # to do remove dns entry
        message = {
            "device":self._group['Cores'][0]['name'],
        }

        response =  self._iot_data_plane.publish(
            topic='d1/remove/'+self._group['Cores'][0]['name'],
            payload= json.dumps(message)
        )



import shutil
import os
from time import sleep
import json
from botocore.exceptions import ClientError
import logging

from .entity import Entity
from .utils import pretty, rinse

MAGIC_DIR = '.gg'  # XXX: better way?

log = logging.getLogger(__name__)


class Lambdas(Entity):
    def __init__(self, group, state):
        super(Lambdas, self).__init__(group, state)
        self.type = 'Lambdas'
        self.name = group['Group']['name'] + '_lambdas'
        self._LAMBDA_ROLE_NAME = "{0}_Lambda_Role".format(self.name)

        self._requirements = ['Group']
        self._gg = Entity._session.client("greengrass")
        self._iam = Entity._session.client("iam")
        self._lambda = Entity._session.client("lambda")

    def _do_create(self):
        functions = []
        self._state.update('Lambdas', {})
        self._state.update('Lambdas.Functions', [])  # XXX do this better.
        for l in self._group['Lambdas']:
            log.info("Creating Lambda function '{0}'".format(l['name']))

            # Use existing role if provided, or create & use default role.
            role_arn = l['role'] if 'role' in l else self._default_lambda_role_arn()
            log.info("Assuming role '{0}'".format(role_arn))

            zf = shutil.make_archive(
                os.path.join(MAGIC_DIR, l['name']), 'zip', l['package'])
            log.debug("Lambda deployment Zipped to '{0}'".format(zf))

            for retry in range(3):
                try:
                    with open(zf, 'rb') as f:
                        lr = self._lambda.create_function(
                            FunctionName=l['name'],
                            Runtime=l['runtime'],
                            Role=role_arn,
                            Handler=l['handler'],
                            Code=dict(ZipFile=f.read()),
                            Environment=dict(Variables=l.get('environment', {})),
                            Publish=True
                        )
                        # Break from retry cycle if lambda is created
                        break
                except ClientError as e:  # Catch the right exception
                    if "The role defined for the function cannot be assumed by Lambda" in str(e):
                        # Function creation immediately after role creation fails with
                        # "The role defined for the function cannot be assumed by Lambda."
                        # See StackOverflow https://goo.gl/eTfqsS
                        log.warning("We hit AWS bug: the role is not yet propagated. "
                                    "Taking 5 sec nap")
                        sleep(5)
                        continue
                    else:
                        raise(e)

            lr['ZipPath'] = zf

            self._state.get('Lambdas.Functions').append(rinse(lr))
            self._state.save()  # This is needed as `append` only modified state in memory.
            log.info("Lambda function '{0}' created".format(lr['FunctionName']))

            # Auto-created alias uses the version of just published function
            alias = self._lambda.create_alias(
                FunctionName=lr['FunctionName'],
                Name=l.get('alias', 'default'),
                FunctionVersion=lr['Version'],
                Description='Created by greengo'
            )
            log.info("Lambda alias created. FunctionVersion:'{0}', Arn:'{1}'".format(
                alias['FunctionVersion'], alias['AliasArn']))

            functions.append({
                'Id': l['name'],
                'FunctionArn': alias['AliasArn'],
                'FunctionConfiguration': l['greengrassConfig']
            })

        log.debug("Function definition list ready:\n{0}".format(pretty(functions)))

        log.info("Creating function definition: '{0}'".format(self.name + '_func_def_1'))
        fd = self._gg.create_function_definition(
            Name=self.name + '_func_def_1',
            InitialVersion={'Functions': functions}
        )
        self._state.update('Lambdas.FunctionDefinition', rinse(fd))

        fd_ver = self._gg.get_function_definition_version(
            FunctionDefinitionId=self._state.get('Lambdas.FunctionDefinition.Id'),
            FunctionDefinitionVersionId=self._state.get('Lambdas.FunctionDefinition.LatestVersion'))

        self._state.update('Lambdas.FunctionDefinition.LatestVersionDetails', rinse(fd_ver))

    def _do_remove(self):

        if not self._state.get('Lambdas.FunctionDefinition'):
            log.warning("Function definition was not created. Moving on...")
        else:
            fd_name = self._state.get('Lambdas.FunctionDefinition.Name')
            fd_id = self._state.get('Lambdas.FunctionDefinition.Id')
            log.info("Deleting function definition '{0}' Id='{1}".format(fd_name, fd_id))
            self._gg.delete_function_definition(FunctionDefinitionId=fd_id)
            self._state.remove('Lambdas.FunctionDefinition')

        if not self._state.get('Lambdas.LambdaRole'):
            log.warning("Lambda Role was not created. Moving on...")
        else:
            log.info("Deleting default lambda role '{0}'".format(self._LAMBDA_ROLE_NAME))
            self._remove_default_lambda_role()
            self._state.remove('Lambdas.LambdaRole')

        if not self._state.get('Lambdas.Functions'):
            log.warning("Lambda Functions were not created. Moving on...")
        else:
            for l in self._state.get('Lambdas.Functions'):
                log.info("Deleting Lambda function '{0}'".format(l['FunctionName']))
                self._lambda.delete_function(FunctionName=l['FunctionName'])
                try:
                    os.remove(l['ZipPath'])
                except OSError as e:
                    log.warning("Failed to remove local Lambda zip package: {}".format(e))

            self._state.remove('Lambdas.Functions')

    def update_lambda(self, lambda_name):
        functions = self._state.get('Lambdas.Functions')
        if not functions:
            log.info("No lambda functions created. Create first...")
            return

        lr = next((lr for lr in functions if lr['FunctionName'] == lambda_name), None)
        if not lr:
            log.error("No lambda function '{0}' found.".format(lambda_name))
            return

        l = next((l for l in self._group['Lambdas'] if l['name'] == lambda_name), None)
        if not l:
            log.error("No definition for lambda function '{0}'.".format(lambda_name))
            return

        log.info("Updating lambda function code for '{0}'".format(lr['FunctionName']))

        zf = shutil.make_archive(
            os.path.join(MAGIC_DIR, l['name']), 'zip', l['package'])
        log.debug("Lambda deployment Zipped to '{0}'".format(zf))

        with open(zf, 'rb') as fd:
            lr_updated = self._lambda.update_function_code(
                FunctionName=l['name'],
                ZipFile=fd.read(),
                Publish=True
            )

        fnew = [rinse(lr_updated) if f['FunctionName'] == lambda_name else f for f in functions]
        self._state.update('Lambdas.Functions', fnew)

        log.info("Lambda function '{0}' updated".format(lr['FunctionName']))

        log.info("Updating alias '{0}'...".format(l.get('alias', 'default')))
        alias = self._lambda.update_alias(
            FunctionName=lr['FunctionName'],
            Name=l.get('alias', 'default'),
            FunctionVersion=lr['Version']
        )

        log.info("Lambda alias updated. FunctionVersion:'{0}', Arn:'{1}'".format(
            alias['FunctionVersion'], alias['AliasArn']))
        # TODO: save alias? If so, where in state?
        # If the alias name changed in group,
        # then LambdaDefinitions should also be updated.

        log.info("Lambdas function {0} updated OK!".format(lambda_name))

    def _default_lambda_role_arn(self):
        # TODO(XXX): Refactor, merge with _create_default_lambda_role;
        #            consider not messing with state here, move it up.
        if self._state.get('Lambdas.LambdaRole'):
            log.info("Default lambda role '{0}' already creted, RoleId={1} ".format(
                self._LAMBDA_ROLE_NAME, self._state.get('LambdaRole.Role.RoleId')))
        else:
            try:
                role = self._create_default_lambda_role()
            except ClientError as e:
                if e.response['Error']['Code'] == 'EntityAlreadyExists':
                    role = self._iam.get_role(RoleName=self._LAMBDA_ROLE_NAME)
                    log.warning("Role {0} already exists, reusing.".format(self._LAMBDA_ROLE_NAME))
                else:
                    raise e

            self._state.update('Lambdas.LambdaRole', rinse(role))
        return self._state.get('Lambdas.LambdaRole.Role.Arn')

    def _create_default_lambda_role(self):
        # TODO: redo as template and read from definition .yaml
        log.info("Creating default lambda role '{0}'".format(self._LAMBDA_ROLE_NAME))
        role_policy_document = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {
                        "Service": "lambda.amazonaws.com"
                    },
                    "Action": "sts:AssumeRole"

                }
            ]
        }

        role = self._iam.create_role(
            RoleName=self._LAMBDA_ROLE_NAME,
            AssumeRolePolicyDocument=json.dumps(role_policy_document)
        )

        log.info("Creating lambda role policy '{0}'".format(
            self._LAMBDA_ROLE_NAME + "_Policy"))
        inline_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "logs:CreateLogGroup",
                        "logs:CreateLogStream",
                        "logs:PutLogEvents"
                    ],
                    "Resource": "arn:aws:logs:*:*:*"
                }
            ]
        }

        self._iam.put_role_policy(
            RoleName=self._LAMBDA_ROLE_NAME,
            PolicyName=self._LAMBDA_ROLE_NAME + "_Policy",
            PolicyDocument=json.dumps(inline_policy))

        return role

    def _remove_default_lambda_role(self):
        for p in self._iam.list_role_policies(RoleName=self._LAMBDA_ROLE_NAME)['PolicyNames']:
            log.info("Deleting lambda role policy '{0}'".format(p))
            self._iam.delete_role_policy(RoleName=self._LAMBDA_ROLE_NAME, PolicyName=p)

        log.info("Deleting default lambda role '{0}'".format(self._LAMBDA_ROLE_NAME))
        self._iam.delete_role(RoleName=self._LAMBDA_ROLE_NAME)



import os
import errno
import logging
import json

from . import utils

log = logging.getLogger(__name__)


class State(object):

    def __init__(self, file):
        # Entities map: { entityName: entity }
        self._entities = {}
        self._state = {}
        self._file = file
        if file:
            self.load()

    def load(self):
        if not os.path.exists(self._file):
            log.debug("Group state file {0} not found, assume new group.".format(self._file))
        else:
            log.debug("Loading group state from {0}".format(self._file))
            with open(self._file, 'r') as f:
                self._state = json.load(f)

    def save(self):
        if not self._file:
            return
        try:
            with open(self._file, 'w') as f:
                json.dump(self._state, f, indent=2,
                          separators=(',', ': '), sort_keys=True, default=str)
                # log.debug("Updated group state in state file '{0}'".format(self._file))
        except IOError as e:
            # Assume we miss the directory... Create it and try again
            if e.errno != errno.ENOENT:
                raise e
            utils.mkdir(os.path.dirname(self._file))
            self.save()

    def __str__(self):
        if not self._state:
            return "State not created."
        return utils.pretty(self._state)

    def exists(self):
        return bool(self._state)

    def update(self, key, body):
        '''
        Updates nested keys using dot as separator, like 'foo.bar.buz'.
        Creates missing keys as nessessary.
        '''
        tree = self._state
        prev = None
        for k in key.split('.'):
            if prev is not None:
                tree = tree.setdefault(prev, {})
            prev = k

        tree[prev] = body
        self.save()

    def entities(self):
        ''' Get list of entities - top level keys '''
        return self._state.keys()

    def get(self, key=None, default=None):
        '''
        Get value. Key may be nested using dot as separator, like 'foo.bar.buz'.
        '''
        branch = self._state
        if key:
            for k in key.split('.'):
                branch = branch.get(k, default)
                if branch == default:
                    break

        return branch

    def remove(self, key=None):
        if key:
            try:
                (path, dot, k) = key.rpartition('.')
                if path:
                    branch = self.get(path)
                    if not branch or type(branch) is not dict:
                        raise KeyError
                    branch.pop(k)
                else:
                    self._state.pop(key)
                self.save()
            except KeyError:
                log.warning("Can not remove key '{}' from State - missing".format(key))
        else:
            self._state = {}
            if self._file:
                try:
                    os.remove(self._file)
                except OSError:
                    log.warning("State file not removed (missing?): {}".format(self._file))



import logging

from .utils import pretty, rinse
from .entity import Entity

log = logging.getLogger(__name__)


class Resources(Entity):

    def __init__(self, group, state):
        super(Resources, self).__init__(group, state)
        self.type = 'Resources'
        self.name = group['Group']['name'] + '_resources'

        self._requirements = ['Group']
        self._gg = Entity._session.client("greengrass")

    def _do_create(self):
        log.debug("Preparing resources list...")
        res = []
        for r in self._group['Resources']:
            # Convert from a simplified form to AWS API input
            resource = dict(Name=r.pop('Name'), Id=r.pop('Id'))
            resource['ResourceDataContainer'] = r
            res.append(resource)

        log.debug("Resources list is ready:\n{0}".format(pretty(res)))

        log.info("Creating resource definition: '{0}'".format(self.name))
        res_def = rinse(self._gg.create_resource_definition(
            Name=self.name,
            InitialVersion={'Resources': res}
        ))

        self._state.update('Resources', res_def)

        res_def_ver = rinse(self._gg.get_resource_definition_version(
            ResourceDefinitionId=self._state.get('Resources.Id'),
            ResourceDefinitionVersionId=self._state.get('Resources.LatestVersion')
        ))

        self._state.update('Resources.LatestVersionDetails', res_def_ver)

    def _do_remove(self):
        log.debug("Deleting resources definition '{0}' Id='{1}".format(
            self.name, self._state.get('Resources.Id')))
        self._gg.delete_resource_definition(
            ResourceDefinitionId=self._state.get('Resources.Id'))

        self._state.remove(self.type)



from setuptools import setup
from codecs import open
from os import path
from greengo import __version__

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='gud',
    version=__version__,
    description='AWS Greengrass file-based setup tool based on greengo by d',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['AWS', 'IoT', 'Greengrass', 'Lambda'],
    url='https://gitlab.com/doopnz/gud',
    author='Dmitri Zimine modified by James Whiteman',
    author_email='james.whiteman@gmail.com',
    license='MIT',
    packages=['greengo'],
    install_requires=[
        'fire',
        'boto3',
        'botocore',
        'pyyaml'
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'mock'],
    entry_points={
        'console_scripts': ['gud=greengo.greengo:main'],
    },
    zip_safe=False
)
